################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from io import StringIO

import dimod
import numpy as np
import pytest

from orquestra.opt.bqm.io import (
    bqm_from_serializable,
    bqm_to_serializable,
    load_qubo,
    load_sampleset,
    save_qubo,
    save_sampleset,
)


class TestConvertingBQMToSerializable:
    def test_all_linear_coefficients_are_stored(self):
        bqm = dimod.BinaryQuadraticModel(
            {0: 1, 1: 2, 2: 3},
            {(1, 2): 0.5, (1, 0): 0.7, (0, 2): 0.9},
            -10,
            vartype=dimod.BINARY,
        )

        serializable = bqm_to_serializable(bqm)

        assert set(serializable["linear"]) == {(0, 1), (1, 2), (2, 3)}

    def test_all_quadratic_coefficients_are_stored(self):
        bqm = dimod.BinaryQuadraticModel(
            {0: 0.5, 2: -2.0, 3: 3},
            {(2, 1): 0.5, (1, 0): 0.4, (0, 3): -0.1},
            -5,
            vartype=dimod.BINARY,
        )

        serializable = bqm_to_serializable(bqm)

        assert set(serializable["quadratic"]) == {
            (1, 2, 0.5),
            (0, 1, 0.4),
            (0, 3, -0.1),
        }

    @pytest.mark.parametrize("offset", [-5, 10, 0])
    def test_offset_is_stored(self, offset):
        bqm = dimod.BinaryQuadraticModel(
            {0: 0.5, 2: -2.0, 3: 3},
            {(2, 1): 0.5, (1, 0): 0.4, (0, 3): -0.1},
            offset,
            vartype=dimod.BINARY,
        )

        serializable = bqm_to_serializable(bqm)

        assert serializable["offset"] == offset

    @pytest.mark.parametrize(
        "vartype, expected_output_vartype",
        [
            ("SPIN", "SPIN"),
            ("BINARY", "BINARY"),
            (dimod.BINARY, "BINARY"),
            (dimod.SPIN, "SPIN"),
        ],
    )
    def test_vartype_is_stored(self, vartype, expected_output_vartype):
        bqm = dimod.BinaryQuadraticModel(
            {0: 0.5, 2: -2.0, 3: 3},
            {(2, 1): 0.5, (1, 0): 0.4, (0, 3): -0.1},
            vartype=vartype,
        )

        serializable = bqm_to_serializable(bqm)

        assert serializable["vartype"] == expected_output_vartype


class TestConstructingBQMFromSerializable:
    def test_all_linear_coefficients_are_loaded(self):
        bqm_dict = {
            "linear": [(0, 2.0), (2, 0.5), (1, -1.0)],
            "quadratic": [(0, 1, 1.2), (1, 2, 4.0)],
            "offset": 0.5,
            "vartype": "SPIN",
        }

        bqm = bqm_from_serializable(bqm_dict)
        assert bqm.linear == {0: 2.0, 2: 0.5, 1: -1.0}

    def test_all_quadratic_coefficients_are_loaded(self):
        bqm_dict = {
            "linear": [(0, 1.0), (1, 2.0), (2, 0.5)],
            "quadratic": [(0, 1, 2.1), (1, 2, 4.0), (1, 3, -1.0)],
            "offset": 0.1,
            "vartype": "BINARY",
        }

        bqm = bqm_from_serializable(bqm_dict)

        assert bqm.quadratic == {(0, 1): 2.1, (1, 2): 4.0, (1, 3): -1.0}

    @pytest.mark.parametrize("offset", [0.1, 2, -3.5])
    def test_offset_is_loaded(self, offset):
        bqm_dict = {
            "linear": [(0, 1.0), (1, 2.0), (2, 0.5)],
            "quadratic": [(0, 1, 2.1), (1, 2, 4.0), (1, 3, -1.0)],
            "offset": offset,
            "vartype": "BINARY",
        }

        bqm = bqm_from_serializable(bqm_dict)

        assert bqm.offset == offset

    @pytest.mark.parametrize(
        "vartype, expected_bqm_vartype",
        [("SPIN", dimod.SPIN), ("BINARY", dimod.BINARY)],
    )
    def test_vartype_is_set_correctly(self, vartype, expected_bqm_vartype):
        bqm_dict = {
            "linear": [(0, 1.0), (2, 2.0), (1, 0.5)],
            "quadratic": [(0, 1, 1), (1, 2, 4.0), (1, 3, 1e-2)],
            "offset": 0.5,
            "vartype": vartype,
        }

        bqm = bqm_from_serializable(bqm_dict)

        assert bqm.vartype == expected_bqm_vartype


def test_loading_saved_qubo_gives_the_same_qubo():
    qubo = dimod.BinaryQuadraticModel(
        {0: 0.5, 2: -2.0, 3: 3},
        {(2, 1): 0.5, (1, 0): 0.4, (0, 3): -0.1},
        -5,
        vartype="BINARY",
    )

    output_file = StringIO()

    save_qubo(qubo, output_file)
    # Move to the beginning of the file
    output_file.seek(0)
    new_qubo = load_qubo(output_file)

    assert qubo == new_qubo


def test_loading_qubo_with_complex_variables():
    qubo = dimod.BinaryQuadraticModel(
        {(0, 0): 1, (1, 0): -1, (2, 0): 0.5},
        {((0, 0), (1, 0)): 0.5, ((1, 0), (2, 0)): 1.5},
        42,
        vartype="SPIN",
    )

    output_file = StringIO()

    save_qubo(qubo, output_file)
    # Move to the beginning of the file
    output_file.seek(0)
    new_qubo = load_qubo(output_file)

    assert qubo == new_qubo


def test_loading_saved_sampleset_gives_the_same_sampleset():
    sampleset = dimod.SampleSet.from_samples(np.ones(5, dtype="int8"), "BINARY", 0)

    output_file = StringIO()

    save_sampleset(sampleset, output_file)
    # Move to the beginning of the file
    output_file.seek(0)
    new_sampleset = load_sampleset(output_file)

    assert sampleset == new_sampleset



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import numpy as np
import pytest

from orquestra.opt.api.optimizer_test import OptimizerTests
from orquestra.opt.optimizers.search_points_optimizer import SearchPointsOptimizer


@pytest.fixture(
    params=[
        [
            np.array([0, 0]),
            np.array([1, 1]),
            np.array([0, 1]),
            np.array([1.3, 0]),
            np.array([0.000001, -1.3]),
        ],
        np.array(
            [
                np.array([0, 0]),
                np.array([1, 1]),
                np.array([0, 1]),
                np.array([1.3, 0]),
                np.array([0.000001, -1.3]),
            ]
        ),
        [
            np.array([1, 1]),
            np.array([0, 0]),
        ],
    ]
)
def parameter_values_list(request):
    return request.param


@pytest.fixture(params=[True, False])
def keep_history(request):
    return request.param


class TestSearchPointsOptimizer(OptimizerTests):
    @pytest.fixture()
    def optimizer(self, parameter_values_list):
        return SearchPointsOptimizer(parameter_values_list=parameter_values_list)

    def test_assertion_raised_when_no_points_passed(self):
        with pytest.raises(AssertionError):
            SearchPointsOptimizer(parameter_values_list=[])



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from functools import partial
from typing import Callable, List, Tuple, cast

import networkx as nx
import numpy as np
from orquestra.quantum.measurements import Measurements, expectation_values_to_real
from orquestra.quantum.openfermion import IsingOperator, change_operator_type
from orquestra.quantum.openfermion.ops.operators.qubit_operator import QubitOperator
from orquestra.quantum.openfermion.utils import count_qubits
from orquestra.quantum.utils import dec2bin


# This is the only function in this file in the Public API
def solve_problem_by_exhaustive_search(
    hamiltonian: QubitOperator,
) -> Tuple[float, List[Tuple[int, ...]]]:
    """
    Solves any QAOA cost hamiltonian using exhaustive search.

    It searches for degeneracy and returns all the best solutions if more than one
    exists.

    Args:
        hamiltonian: cost hamiltonian

    Returns:
        float: value of the best solution
        List[Tuple[int]]: list of solutions which correspond to the best value, each
            solution is a tuple of ints.
    """
    cost_function = _evaluate_solution_for_hamiltonian

    return _solve_bitstring_problem_by_exhaustive_search(
        partial(cost_function, hamiltonian=hamiltonian),
        num_nodes=count_qubits(hamiltonian),
    )


def solve_graph_problem_by_exhaustive_search(
    graph: nx.Graph,
    cost_function: Callable[[Tuple[int], nx.Graph], float],
) -> Tuple[float, List[Tuple[int, ...]]]:
    """
    Solves given graph problem using exhaustive search.

    It searches for degeneracy and returns all the best solutions if more than one
    exists.

    Args:
        graph: graph for which we want to solve the problem
        cost_function: function which calculates the cost of solution of a given
            problem.

    Returns:
        float: value of the best solution
        List[Tuple[int]]: list of solutions which correspond to the best value, each
            solution is a tuple of ints.
    """
    num_nodes = graph.number_of_nodes()
    return _solve_bitstring_problem_by_exhaustive_search(
        partial(cost_function, graph=graph), num_nodes=num_nodes
    )


def _solve_bitstring_problem_by_exhaustive_search(
    cost_function: Callable[[Tuple[int, ...]], float],
    num_nodes: int,
) -> Tuple[float, List[Tuple[int, ...]]]:
    """
    Solves given cost function of a graph problem using exhaustive search.

    It searches for degeneracy and returns all the best solutions if more than one
    exists.

    Args:
        cost_function: function which calculates the cost of solution of a given
            problem. Callable that takes a bitstring solution and outputs cost value.
        num_nodes: number of nodes of the graph for which we want to solve the problem

    Returns:
        float: value of the best solution
        List[Tuple[int]]: list of solutions which correspond to the best value, each
            solution is a tuple of ints.
    """
    solutions_list = []

    best_value = np.inf

    for i in range(2**num_nodes):
        trial_solution: Tuple[int, ...] = tuple(dec2bin(i, num_nodes))
        current_value = cost_function(trial_solution)
        if current_value == best_value:
            solutions_list.append(trial_solution)
        if current_value < best_value:
            best_value = current_value
            solutions_list = [trial_solution]

    return best_value, solutions_list


def evaluate_solution(
    solution: Tuple[int],
    graph: nx.Graph,
    get_hamiltonian: Callable[[nx.Graph], QubitOperator],
) -> float:
    """Evaluate expectation value of hamiltonian for given solution of a graph problem.

    Args:
        solution: solution to a problem as a tuple of bits
        graph: a graph for which we want to solve the problem
        get_hamiltonian: function which translates graph into a Hamiltonian representing
            a problem.

    Returns:
        float: value of a solution.
    """
    if len(graph.nodes) != len(solution):
        raise ValueError("Length of solution must match size of the graph.")
    if any(el not in [0, 1] for el in solution):
        raise ValueError("Solution must consist of either 0s or 1s.")
    return _evaluate_solution_for_hamiltonian(solution, get_hamiltonian(graph))


def _evaluate_solution_for_hamiltonian(
    solution: Tuple[int], hamiltonian: QubitOperator
) -> float:
    """Evaluates a solution of a hamiltonian by its calculating expectation value.

    Args:
        solution: solution to a problem as a tuple of bits
        hamiltonian: a Hamiltonian representing a problem.

    Returns:
        float: value of a solution.
    """
    hamiltonian = change_operator_type(hamiltonian, IsingOperator)

    expectation_values = expectation_values_to_real(
        Measurements([solution]).get_expectation_values(
            cast(IsingOperator, hamiltonian)
        )
    )
    return sum(expectation_values.values)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Test case prototypes that can be used in other projects.

Note that this file won't be executed on its own by pytest.
You need to define your own test cases that inherit from the ones defined here.
"""


from typing import Callable

import numpy as np
import pytest

from ..gradients import finite_differences_gradient
from ..history.recorder import recorder
from .cost_function import CostFunction
from .functions import FunctionWithGradient
from .optimizer import NestedOptimizer

MANDATORY_OPTIMIZATION_RESULT_FIELDS = ("nfev", "nit", "opt_value", "opt_params")


class OptimizerTests(object):
    """Base class for optimizers tests.

    How to use:
    1. Inherit this class (remember to start name of the class with "Test"
    2. In the same module define fixture called "optimizer".

    Basic usage pattern:

    @pytest.fixture
    def optimizer():
        return MyOptimizer()


    class TestMyOptimizer(OptimizerTests): # Inherits all tests from this class
         def test_some_new_feature(self, optimizer): # new test
             ....

    Notice that the `optimizer` fixture can be parametrized if you wish to
    perform tests for various configurations of your optimizer.
    """

    @pytest.fixture
    def sum_x_squared(self):
        return lambda x: sum(x**2.0)

    @pytest.fixture
    def rosenbrock_function(self):
        """The Rosenbrock function"""
        return lambda x: sum(
            100.0 * (x[1:] - x[:-1] ** 2.0) ** 2.0 + (1 - x[:-1]) ** 2.0
        )

    def test_optimizer_succeeds_with_optimizing_rosenbrock_function(
        self, optimizer, rosenbrock_function, keep_history
    ):
        cost_function = FunctionWithGradient(
            rosenbrock_function, finite_differences_gradient(rosenbrock_function)
        )

        results = optimizer.minimize(
            cost_function, initial_params=np.array([0, 0]), keep_history=keep_history
        )
        assert results.opt_value == pytest.approx(0, abs=1e-4)
        assert results.opt_params == pytest.approx(np.ones(2), abs=1e-3)

        assert all(field in results for field in MANDATORY_OPTIMIZATION_RESULT_FIELDS)

        assert "history" in results or not keep_history
        assert "gradient_history" in results or not keep_history

    def test_optimizer_succeeds_with_optimizing_sum_of_squares_function(
        self, optimizer, sum_x_squared, keep_history
    ):

        cost_function = FunctionWithGradient(
            sum_x_squared, finite_differences_gradient(sum_x_squared)
        )

        results = optimizer.minimize(
            cost_function, initial_params=np.array([1, -1]), keep_history=keep_history
        )

        assert results.opt_value == pytest.approx(0, abs=1e-5)
        assert results.opt_params == pytest.approx(np.zeros(2), abs=1e-4)

        assert all(field in results for field in MANDATORY_OPTIMIZATION_RESULT_FIELDS)

        assert "history" in results or not keep_history
        assert "gradient_history" in results or not keep_history

    def test_optimizer_succeeds_on_cost_function_without_gradient(
        self, optimizer, sum_x_squared, keep_history
    ):
        cost_function = sum_x_squared

        results = optimizer.minimize(
            cost_function, initial_params=np.array([1, -1]), keep_history=keep_history
        )
        assert results.opt_value == pytest.approx(0, abs=1e-5)
        assert results.opt_params == pytest.approx(np.zeros(2), abs=1e-4)

        assert all(field in results for field in MANDATORY_OPTIMIZATION_RESULT_FIELDS)

        assert "history" in results or not keep_history
        assert "gradient_history" not in results

    def test_optimizer_records_history_if_keep_history_is_true(
        self, optimizer, sum_x_squared
    ):

        # To check that history is recorded correctly, we wrap cost_function
        # with a recorder. Optimizer should wrap it a second time and
        # therefore we can compare two histories to see if they agree.
        cost_function = recorder(sum_x_squared)

        result = optimizer.minimize(cost_function, np.array([-1, 1]), keep_history=True)

        for result_history_entry, cost_function_history_entry in zip(
            result.history, cost_function.history
        ):
            assert (
                result_history_entry.call_number
                == cost_function_history_entry.call_number
            )
            assert np.allclose(
                result_history_entry.params, cost_function_history_entry.params
            )
            assert np.isclose(
                result_history_entry.value, cost_function_history_entry.value
            )

    def test_gradients_history_is_recorded_if_keep_history_is_true(
        self, optimizer, sum_x_squared
    ):
        # To check that history is recorded correctly, we wrap cost_function
        # with a recorder. Optimizer should wrap it a second time and
        # therefore we can compare two histories to see if they agree.
        cost_function = recorder(
            FunctionWithGradient(
                sum_x_squared, finite_differences_gradient(sum_x_squared)
            )
        )

        result = optimizer.minimize(cost_function, np.array([-1, 1]), keep_history=True)
        assert len(result.gradient_history) == len(cost_function.gradient.history)

        for result_history_entry, cost_function_history_entry in zip(
            result.gradient_history, cost_function.gradient.history
        ):
            assert (
                result_history_entry.call_number
                == cost_function_history_entry.call_number
            )
            assert np.allclose(
                result_history_entry.params, cost_function_history_entry.params
            )
            assert np.allclose(
                result_history_entry.value, cost_function_history_entry.value
            )

    def test_optimizer_does_not_record_history_if_keep_history_is_set_to_false(
        self, optimizer, sum_x_squared
    ):
        result = optimizer.minimize(
            sum_x_squared, np.array([-2, 0.5]), keep_history=False
        )

        assert result.history == []

    def test_optimizer_does_not_record_history_by_default(
        self, optimizer, sum_x_squared
    ):
        result = optimizer.minimize(sum_x_squared, np.array([-2, 0.5]))

        assert result.history == []


"""Contracts for instances of the NestedOptimizer base class that can be used in
other projects.

Usage:

    from orquestra.opt.api.optimizer_test import NESTED_OPTIMIZER_CONTRACTS

    @pytest.mark.parametrize("contract", NESTED_OPTIMIZER_CONTRACTS)
    def test_nestedoptimizer_contract(contract):
        optimizer = MockNestedOptimizer(inner_optimizer=MockOptimizer(), n_iters=5)
        assert contract(optimizer)
"""


def _validate_nested_optimizer_records_history_if_keep_history_is_true(
    optimizer: NestedOptimizer,
    cost_function_factory: Callable[..., CostFunction],
    initial_params: np.ndarray,
):
    result = optimizer.minimize(
        cost_function_factory, initial_params, keep_history=True
    )
    return len(result.history) != 0


def _validate_nested_optimizer_records_gradient_history_if_keep_history_is_true(
    optimizer: NestedOptimizer,
    cost_function_factory: Callable[..., CostFunction],
    initial_params: np.ndarray,
):
    def cost_function_with_gradients_factory(*args, **kwargs):
        cost_function = cost_function_factory(*args, **kwargs)
        return FunctionWithGradient(
            cost_function, finite_differences_gradient(cost_function)
        )

    result = optimizer.minimize(
        cost_function_with_gradients_factory, initial_params, keep_history=True
    )
    return hasattr(result, "gradient_history")


def _validate_nested_optimizer_does_not_record_history_if_keep_history_is_false(
    optimizer: NestedOptimizer,
    cost_function_factory: Callable[..., CostFunction],
    initial_params: np.ndarray,
):
    result = optimizer.minimize(
        cost_function_factory, initial_params, keep_history=False
    )
    return len(result.history) == 0


def _validate_nested_optimizer_does_not_record_history_by_default(
    optimizer: NestedOptimizer,
    cost_function_factory: Callable[..., CostFunction],
    initial_params: np.ndarray,
):

    result = optimizer.minimize(cost_function_factory, initial_params)
    return result.history == []


def _validate_nested_optimizer_returns_all_the_mandatory_fields_in_results(
    optimizer: NestedOptimizer,
    cost_function_factory: Callable[..., CostFunction],
    initial_params: np.ndarray,
):
    result = optimizer.minimize(cost_function_factory, initial_params)
    return all(field in result for field in MANDATORY_OPTIMIZATION_RESULT_FIELDS)


NESTED_OPTIMIZER_CONTRACTS = [
    _validate_nested_optimizer_records_history_if_keep_history_is_true,
    _validate_nested_optimizer_records_gradient_history_if_keep_history_is_true,
    _validate_nested_optimizer_does_not_record_history_if_keep_history_is_false,
    _validate_nested_optimizer_does_not_record_history_by_default,
    _validate_nested_optimizer_returns_all_the_mandatory_fields_in_results,
]



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from orquestra.opt.api import function_with_gradient, has_store_artifact_param
from orquestra.opt.gradients import finite_differences_gradient


class TestAddingGradient:
    class TestFunctionStoringArtifacts:
        def test_makes_a_callable_that_stores_artifacts(self):
            def _test_function(params, store_artifact=None):
                if store_artifact:
                    store_artifact("x", params[0])
                return (params**2).sum()

            function = function_with_gradient(
                _test_function, finite_differences_gradient(_test_function)
            )
            assert has_store_artifact_param(function)

    class TestFunctionNotStoringArtifacts:
        def test_makes_a_callable_not_storing_artifacts(self):
            def _test_function(params):
                return (params**2).sum()

            function = function_with_gradient(
                _test_function, finite_differences_gradient(_test_function)
            )
            assert not has_store_artifact_param(function)
